################################################################################
# INSTALL PACKAGES AND CALL LIBRARIES
################################################################################
if(!require(glmnet)){ install.packages("glmnet"); }
if(!require(hqreg)){ install.packages("hqreg"); }
library(glmnet);
library(hqreg);
source("code_kernel.r");

################################################################################
# BASIC SIMULATION SETTINGS
################################################################################
n = 500; 
p = 50;
L = 1;
DGP = 1;
taulist = 10:40/50;
num_mc = 500;
num_boot = 500;
max_parallel = 3;

################################################################################
# APPROXIMATE TRUE PARAMETER VALUES
################################################################################
source("code_approximate_true_parameter_values.r");
options(scipen=999);
options(digits=3);
cbind( taulist, trueQY, truefY, truetheta, trueUQPE ); 
Sys.sleep(2);









################################################################################
# BEGIN MONTE CARLO
################################################################################
mc_thetahat=0;
mc_thetaSqErr=0;
mc_QtauYhat=0;
mc_QYSqErr=0;
mc_fYhat=0;
mc_fYQqErr=0;
mc_UQPEhat=0;
mc_UQPESqErr=0;
mc_boot_uniform_80cover=0;
mc_boot_uniform_90cover=0;
mc_boot_uniform_95cover=0;
mc_boot_uniform_99cover=0;
mc_boot_pointwise_80cover=0;
mc_boot_pointwise_90cover=0;
mc_boot_pointwise_95cover=0;
mc_boot_pointwise_99cover=0;
mc_analytic_pointwise_80cover=0;
mc_analytic_pointwise_90cover=0;
mc_analytic_pointwise_95cover=0;
mc_analytic_pointwise_99cover=0;
for( iter_MC in 1:num_mc ){

################################################################################
# GENERATE DATA
################################################################################
source("code_generate_data.r");

################################################################################
# ESTIMATION
################################################################################
m0 = matrix(0,n,length(taulist));
m1 = matrix(0,n,length(taulist));
ell = matrix(0,n,1);

################################################################################
# SAMPLE SPLITTING
################################################################################
#   fold = ceiling(sample(n,replace=FALSE)/(n/L))
QtauYhat = matrix(0,1,length(taulist));
thetahat = matrix(0,1,length(taulist));

#   for( current_l in 1:L ){
 score_indices = 1:n # which(fold==current_l);
 nuisance_indices = 1:n # which(fold!=current_l);
 current_n = length( nuisance_indices );

  # ESTIMATE Ell ################################################################
#    curr_x = cbind( X2p[nuisance_indices,], X2p[nuisance_indices,]^2, X2p[nuisance_indices,]^3 );
#    curr_n = dim(curr_x)[1];
#    curr_p = dim(curr_x)[2];
#    curr_y = X1[nuisance_indices];
#    lambdamin =  1.1 * qnorm(1 - 1/(log(curr_p)*max(c(curr_n,curr_p)))) * curr_n^(-0.5);
#  
#  #  Psihat = diag( colMeans( matrix(0.25*0.75,length(curr_y),curr_p) * curr_x^2 )^0.5 );
#  #  lassoresultQ25 = hqreg( curr_x %*% solve(Psihat), curr_y, method = "quantile", tau = 0.25, preprocess = "standardize" );
#  #  lassoresultQ75 = hqreg( curr_x %*% solve(Psihat), curr_y, method = "quantile", tau = 0.75, preprocess = "standardize" );
#  
#    Psihat = colMeans( matrix(0.25*0.75,length(curr_y),curr_p) * curr_x^2 )^0.5;
#    lassoresultQ25 = hqreg( curr_x, curr_y, method = "quantile", tau = 0.25, penalty.factor = Psihat );
#    lassoresultQ75 = hqreg( curr_x, curr_y, method = "quantile", tau = 0.75, penalty.factor = Psihat );
#  
#    for( score_index in score_indices ){
#     curr_newx = matrix( c( X2p[score_index,], X2p[score_index,]^2, X2p[score_index,]^3 ), 1, curr_p );
#  
#  #   Q25 = predict(lassoresultQ25, curr_newx %*% solve(Psihat), lambda = lambdamin );
#  #   Q75 = predict(lassoresultQ75, curr_newx %*% solve(Psihat), lambda = lambdamin );
#  
#     Q25 = predict(lassoresultQ25, curr_newx, lambda = lambdamin, penalty.factor = Psihat );
#     Q75 = predict(lassoresultQ75, curr_newx, lambda = lambdamin, penalty.factor = Psihat );
#  
#     muell = (Q25 + Q75) / 2; 
#     sigmaell = (Q25 - Q75) / (pnorm(0.25) - pnorm(0.75)); 
#   
#     curr_ell = - ( X1[score_index] - muell ) / sigmaell^2;
#     ell[score_index,1] = curr_ell;
#    }
#    rm(curr_x); rm(curr_y); rm(Psihat); rm(lassoresultQ25); rm(lassoresultQ75); gc();

 for( tdx in 1:length(taulist) ){
  tau = taulist[tdx];

  # ESTIMATE QtauYhat ###########################################################
  QtauYhatell = quantile(Y[],p=tau);

  # ESTIMATE m0 #################################################################
  curr_x = cbind( X1[nuisance_indices], X1[nuisance_indices]^2, X1[nuisance_indices]^3, X2p[nuisance_indices,], X2p[nuisance_indices,]^2, X2p[nuisance_indices,]^3 );
  curr_n = dim(curr_x)[1];
  curr_p = dim(curr_x)[2];
  curr_y = 1 * ( Y[nuisance_indices] <= QtauYhatell  );
#    lambdamin =  1.1 * qnorm(1 - 1/(log(curr_p)*max(c(curr_n,curr_p)))) * curr_n^(-0.5);

#  Psihat = diag( colMeans( matrix(curr_y,length(curr_y),curr_p) * curr_x^2 ) );
#  lassoresult = glmnet( curr_x %*% solve(Psihat), curr_y, lambda = lambdamin, family = "binomial", standardize = TRUE, intercept = TRUE );
#  mcheck = cbind(1,curr_x) %*% coef(lassoresult)
#  Psihat = diag( colMeans( matrix(curr_y-mcheck,length(curr_y),curr_p) * curr_x^2 ) );
#  lassoresult = glmnet( curr_x %*% solve(Psihat), curr_y, lambda = lambdamin, family = "binomial", standardize = TRUE, intercept = TRUE );

#    Psihat = colMeans( matrix(curr_y,length(curr_y),curr_p)^2 * curr_x^2 )^0.5;
#    lassoresult = glmnet( curr_x, curr_y, lambda = lambdamin, family = "binomial", intercept = TRUE, penalty.factor = Psihat );
#    mcheck = cbind(1,curr_x) %*% coef(lassoresult)
#    Psihat = colMeans( matrix(curr_y-mcheck,length(curr_y),curr_p)^2 * curr_x^2 )^0.5;
#    lassoresult = glmnet( curr_x, curr_y, lambda = lambdamin, family = "binomial", intercept = TRUE, penalty.factor = Psihat );

#    support = which( as.numeric(coef(lassoresult))[2:(curr_p+1)] | ( 1:curr_p <= 1 ) );
#    selected_x = curr_x[,support];
  selected_x = curr_x
  postlassoresult = glm( curr_y ~ selected_x, family = "binomial" );
  postlasscoef = matrix(0,curr_p+1,1);
  postlasscoef = as.matrix(coef(postlassoresult));

  for( score_index in score_indices ){
   curr_newx = matrix( c( X1[score_index,]^1, X1[score_index,]^2, X1[score_index,]^3, X2p[score_index,], X2p[score_index,]^2, X2p[score_index,]^3 ), 1, curr_p );
   curr_newx_deriv = matrix( c( 1, 2*X1[score_index,], 3*X1[score_index,]^2), 1, 3 );

   FY = plogis( cbind(1,curr_newx) %*% postlasscoef );
   dFYdX1 = dlogis( cbind(1,curr_newx) %*% postlasscoef ) * ( curr_newx_deriv %*% postlasscoef[2:4] )

   m0[score_index,tdx] = FY;
   m1[score_index,tdx] = dFYdX1;
  }
  rm(curr_x); rm(curr_y); rm(Psihat); rm(lassoresult); gc();

#    thetahat_ell = mean( m1[score_indices,tdx] - ell[score_indices,1] * ( 1*(Y[score_indices] <= QtauYhatell) - m0[score_indices,tdx] ) );
  thetahat_ell = mean( m1[score_indices,tdx] );

  QtauYhat[,tdx] = QtauYhatell;
  thetahat[,tdx] = thetahat_ell;
 }# END for ( tdx in 1:length(taulist) ) #######################################
#   }# END for( current_l in 1:L ) #################################################

QtauYhat = colMeans(as.matrix(QtauYhat));
thetahat = colMeans(as.matrix(thetahat));

# ESTIMATE fY ##################################################################
source("code_kernel.r");
fYhat = array(0,length(taulist));
bw = 1.06 * sd(Y) / n^0.2 / n^0.01;
for( tdx in 1:length(taulist) ){
 tau = taulist[tdx];
 fYhat[tdx] = sum( kernel( ( Y - QtauYhat[tdx] ) / bw ) ) / (n * bw);
}

# ESTIMATE UQPE ################################################################
UQPEhat = -1 * thetahat / fYhat;

# ANALYTIC VARIANCE ############################################################
analytic_V_fYhat = fYhat * int_sq_kernel4 / (n * bw);
analytic_V_UQPEhat = thetahat^2 / fYhat^4 * analytic_V_fYhat;
analytic_SD_UQPEhat = analytic_V_UQPEhat^0.5;

################################################################################
# END OF ESTIMATION
################################################################################

################################################################################
# BOOTSTRAP
################################################################################
library(doParallel);
cl<-makeCluster(max_parallel);
registerDoParallel(cl);
bootstrap_results <- foreach( iter_B = 1:num_boot, .combine = rbind )%dopar%{
library(glmnet);
library(hqreg);

eta = rnorm(n,0,1);

################################################################################
# SAMPLE SPLITTING INSIDE BOOTSTRAP
################################################################################
boot_QtauYhat = matrix(0,1,length(taulist));
boot_thetahat = matrix(0,1,length(taulist));

for( current_l in 1:L ){
 score_indices = 1:n # which(fold==current_l);
 nuisance_indices = 1:n # which(fold!=current_l);
 current_n = length( nuisance_indices );

 for( tdx in 1:length(taulist) ){
  tau = taulist[tdx];

  # ESTIMATE QtauYhat ###########################################################
  tau_star = tau + mean( eta[] * (tau - 1*(Y[] <= QtauYhat[tdx])) );
  tau_star = ( tau_star < 0.01 ) * 0.01 + ( tau_star >= 0.01 ) * tau_star;
  tau_star = ( tau_star > 0.99 ) * 0.99 + ( tau_star <= 0.99 ) * tau_star;
  QtauYhatell = quantile(Y[],p=tau_star);

  thetahat_ell = sum( ( eta[score_indices] + 1 ) * ( m1[score_indices,tdx] - ell[score_indices,1] * ( 1*(Y[score_indices] <= QtauYhatell) - m0[score_indices,tdx] ) ) ) / sum( ( eta[score_indices] + 1 ) ) ;

  boot_QtauYhat[current_l,tdx] = QtauYhatell;
  boot_thetahat[current_l,tdx] = thetahat_ell;
 }# END for ( tdx in 1:length(taulist) ) #######################################
}# END for( current_l in 1:L ) #################################################

boot_QtauYhat = colMeans(as.matrix(boot_QtauYhat));
boot_thetahat = colMeans(as.matrix(boot_thetahat));

# ESTIMATE fY ##################################################################
source("code_kernel.r");
boot_fYhat = array(0,length(taulist));
bw = 1.06 * sd(Y) / n^0.2 / n^0.01;
for( tdx in 1:length(taulist) ){
 tau = taulist[tdx];
 boot_fYhat[tdx] = sum( (eta + 1) * kernel( ( Y - QtauYhat[tdx] ) / bw ) ) / (sum(eta + 1) * bw);
}

# ESTIMATE UQPE ################################################################
boot_UQPEhat = -1 * boot_thetahat / boot_fYhat;

boot_UQPEhat;
}# END foreach iter_B #
stopCluster(cl);
################################################################################
# END OF BOOTSTRAP
################################################################################
bootstrap_SD_UQPEhat = apply( bootstrap_results, 2, sd )

################################################################################
# COMPUTE COVERAGE
################################################################################
#sd_boot_results = apply(bootstrap_results,2,sd);
#sd_boot_results = ( apply(bootstrap_results,2,quantile,probs=0.75) - apply(bootstrap_results,2,quantile,probs=0.25) ) / ( qnorm(0.75) - qnorm(0.25) )
#half_lengths_80uniform = quantile( apply( abs( ( bootstrap_results - t(matrix(colMeans(bootstrap_results),length(taulist),num_boot)) ) / t(matrix(sd_boot_results,length(taulist),num_boot)) ), 1, max ), probs = 0.80 );
#half_lengths_90uniform = quantile( apply( abs( ( bootstrap_results - t(matrix(colMeans(bootstrap_results),length(taulist),num_boot)) ) / t(matrix(sd_boot_results,length(taulist),num_boot)) ), 1, max ), probs = 0.90 );
#half_lengths_95uniform = quantile( apply( abs( ( bootstrap_results - t(matrix(colMeans(bootstrap_results),length(taulist),num_boot)) ) / t(matrix(sd_boot_results,length(taulist),num_boot)) ), 1, max ), probs = 0.95 );
#half_lengths_99uniform = quantile( apply( abs( ( bootstrap_results - t(matrix(colMeans(bootstrap_results),length(taulist),num_boot)) ) / t(matrix(sd_boot_results,length(taulist),num_boot)) ), 1, max ), probs = 0.99 );

sd_boot_results = apply(bootstrap_results,2,sd);
#sd_boot_results = ( apply(bootstrap_results,2,quantile,probs=0.75) - apply(bootstrap_results,2,quantile,probs=0.25) ) / ( qnorm(0.75) - qnorm(0.25) )
half_lengths_80uniform = quantile( apply( abs( ( bootstrap_results - t(matrix(UQPEhat,length(taulist),num_boot)) ) / t(matrix(sd_boot_results+0.000000000001,length(taulist),num_boot)) ), 1, max ), probs = 0.80 );
half_lengths_90uniform = quantile( apply( abs( ( bootstrap_results - t(matrix(UQPEhat,length(taulist),num_boot)) ) / t(matrix(sd_boot_results+0.000000000001,length(taulist),num_boot)) ), 1, max ), probs = 0.90 );
half_lengths_95uniform = quantile( apply( abs( ( bootstrap_results - t(matrix(UQPEhat,length(taulist),num_boot)) ) / t(matrix(sd_boot_results+0.000000000001,length(taulist),num_boot)) ), 1, max ), probs = 0.95 );
half_lengths_99uniform = quantile( apply( abs( ( bootstrap_results - t(matrix(UQPEhat,length(taulist),num_boot)) ) / t(matrix(sd_boot_results+0.000000000001,length(taulist),num_boot)) ), 1, max ), probs = 0.99 );

boot_uniform_80CB_lower = UQPEhat - half_lengths_80uniform * sd_boot_results;
boot_uniform_80CB_upper = UQPEhat + half_lengths_80uniform * sd_boot_results;
boot_uniform_90CB_lower = UQPEhat - half_lengths_90uniform * sd_boot_results;
boot_uniform_90CB_upper = UQPEhat + half_lengths_90uniform * sd_boot_results;
boot_uniform_95CB_lower = UQPEhat - half_lengths_95uniform * sd_boot_results;
boot_uniform_95CB_upper = UQPEhat + half_lengths_95uniform * sd_boot_results;
boot_uniform_99CB_lower = UQPEhat - half_lengths_99uniform * sd_boot_results;
boot_uniform_99CB_upper = UQPEhat + half_lengths_99uniform * sd_boot_results;

boot_uniform_80cover = min( boot_uniform_80CB_lower <= trueUQPE & trueUQPE <= boot_uniform_80CB_upper );
boot_uniform_90cover = min( boot_uniform_90CB_lower <= trueUQPE & trueUQPE <= boot_uniform_90CB_upper );
boot_uniform_95cover = min( boot_uniform_95CB_lower <= trueUQPE & trueUQPE <= boot_uniform_95CB_upper );
boot_uniform_99cover = min( boot_uniform_99CB_lower <= trueUQPE & trueUQPE <= boot_uniform_99CB_upper );

boot_pointwise_80CB_lower = UQPEhat + qnorm(0.10)*bootstrap_SD_UQPEhat;
boot_pointwise_80CB_upper = UQPEhat + qnorm(0.90)*bootstrap_SD_UQPEhat;
boot_pointwise_90CB_lower = UQPEhat + qnorm(0.05)*bootstrap_SD_UQPEhat;
boot_pointwise_90CB_upper = UQPEhat + qnorm(0.95)*bootstrap_SD_UQPEhat;
boot_pointwise_95CB_lower = UQPEhat + qnorm(0.025)*bootstrap_SD_UQPEhat;
boot_pointwise_95CB_upper = UQPEhat + qnorm(0.975)*bootstrap_SD_UQPEhat;
boot_pointwise_99CB_lower = UQPEhat + qnorm(0.005)*bootstrap_SD_UQPEhat;
boot_pointwise_99CB_upper = UQPEhat + qnorm(0.995)*bootstrap_SD_UQPEhat;

boot_pointwise_80cover = 1*( boot_pointwise_80CB_lower <= trueUQPE & trueUQPE <= boot_pointwise_80CB_upper );
boot_pointwise_90cover = 1*( boot_pointwise_90CB_lower <= trueUQPE & trueUQPE <= boot_pointwise_90CB_upper );
boot_pointwise_95cover = 1*( boot_pointwise_95CB_lower <= trueUQPE & trueUQPE <= boot_pointwise_95CB_upper );
boot_pointwise_99cover = 1*( boot_pointwise_99CB_lower <= trueUQPE & trueUQPE <= boot_pointwise_99CB_upper );

analytic_pointwise_80CB_lower = UQPEhat + qnorm(0.10)*analytic_SD_UQPEhat;
analytic_pointwise_80CB_upper = UQPEhat + qnorm(0.90)*analytic_SD_UQPEhat;
analytic_pointwise_90CB_lower = UQPEhat + qnorm(0.05)*analytic_SD_UQPEhat;
analytic_pointwise_90CB_upper = UQPEhat + qnorm(0.95)*analytic_SD_UQPEhat;
analytic_pointwise_95CB_lower = UQPEhat + qnorm(0.025)*analytic_SD_UQPEhat;
analytic_pointwise_95CB_upper = UQPEhat + qnorm(0.975)*analytic_SD_UQPEhat;
analytic_pointwise_99CB_lower = UQPEhat + qnorm(0.005)*analytic_SD_UQPEhat;
analytic_pointwise_99CB_upper = UQPEhat + qnorm(0.995)*analytic_SD_UQPEhat;

analytic_pointwise_80cover = 1*( analytic_pointwise_80CB_lower <= trueUQPE & trueUQPE <= analytic_pointwise_80CB_upper );
analytic_pointwise_90cover = 1*( analytic_pointwise_90CB_lower <= trueUQPE & trueUQPE <= analytic_pointwise_90CB_upper );
analytic_pointwise_95cover = 1*( analytic_pointwise_95CB_lower <= trueUQPE & trueUQPE <= analytic_pointwise_95CB_upper );
analytic_pointwise_99cover = 1*( analytic_pointwise_99CB_lower <= trueUQPE & trueUQPE <= analytic_pointwise_99CB_upper );

################################################################################
# DISPLAY RESULTS
################################################################################
rbind(truetheta,thetahat);
rbind(trueQY,QtauYhat);
rbind(truefY,fYhat);
rbind(trueUQPE,UQPEhat, analytic_SD_UQPEhat, bootstrap_SD_UQPEhat);
t(rbind(boot_uniform_95CB_upper,UQPEhat,boot_uniform_95CB_lower));
t(rbind(boot_pointwise_95CB_upper,UQPEhat,boot_pointwise_95CB_lower));
rbind(analytic_pointwise_95CB_upper,UQPEhat,analytic_pointwise_95CB_lower);
rbind(boot_uniform_80cover,boot_uniform_90cover,boot_uniform_95cover,boot_uniform_99cover);
rbind(boot_pointwise_80cover,boot_pointwise_90cover,boot_pointwise_95cover,boot_pointwise_99cover);
rbind(analytic_pointwise_80cover,analytic_pointwise_90cover,analytic_pointwise_95cover,analytic_pointwise_99cover);

################################################################################
# MONTE CARLO UPDATE RESULTS
################################################################################
mc_thetahat=mc_thetahat + thetahat;
mc_thetaSqErr=mc_thetaSqErr + (thetahat-truetheta)^2;
mc_QtauYhat=mc_QtauYhat + QtauYhat;
mc_QYSqErr=mc_QYSqErr + (QtauYhat-trueQY)^2;
mc_fYhat=mc_fYhat + fYhat;
mc_fYQqErr=mc_fYQqErr + (fYhat-truefY)^2;
mc_UQPEhat=mc_UQPEhat + UQPEhat;
mc_UQPESqErr=mc_UQPESqErr + (UQPEhat-trueUQPE)^2;
mc_boot_uniform_80cover=mc_boot_uniform_80cover + boot_uniform_80cover;
mc_boot_uniform_90cover=mc_boot_uniform_90cover + boot_uniform_90cover;
mc_boot_uniform_95cover=mc_boot_uniform_95cover + boot_uniform_95cover;
mc_boot_uniform_99cover=mc_boot_uniform_99cover + boot_uniform_99cover;
mc_boot_pointwise_80cover=mc_boot_pointwise_80cover + boot_pointwise_80cover;
mc_boot_pointwise_90cover=mc_boot_pointwise_90cover + boot_pointwise_90cover;
mc_boot_pointwise_95cover=mc_boot_pointwise_95cover + boot_pointwise_95cover;
mc_boot_pointwise_99cover=mc_boot_pointwise_99cover + boot_pointwise_99cover;
mc_analytic_pointwise_80cover=mc_analytic_pointwise_80cover + analytic_pointwise_80cover;
mc_analytic_pointwise_90cover=mc_analytic_pointwise_90cover + analytic_pointwise_90cover;
mc_analytic_pointwise_95cover=mc_analytic_pointwise_95cover + analytic_pointwise_95cover;
mc_analytic_pointwise_99cover=mc_analytic_pointwise_99cover + analytic_pointwise_99cover;

Sys.sleep(0.1);
print(paste("Monte Carlo Iteration: ",iter_MC,"/",num_mc," 90% Cover = ",mc_boot_uniform_90cover/iter_MC," 95% Cover = ",mc_boot_uniform_95cover/iter_MC,sep=""));
flush.console();
}# END for( iter_MC in 1:num_mc ) #############################################


################################################################################
# SUMMARIZE MONTE CARLO RESULTS
################################################################################
mean_theta = mc_thetahat/iter_MC;
bias_theta = mean_theta - truetheta;
rmse_theta = (mc_thetaSqErr/iter_MC)^0.5;
mean_QY = mc_QtauYhat/iter_MC;
bias_QY = mean_QY - trueQY;
rmse_QY = (mc_QYSqErr/iter_MC)^0.5;
mean_fY = mc_fYhat/iter_MC;
bias_fY = mean_fY - truefY;
rmse_fY = (mc_fYQqErr/iter_MC)^0.5;
mean_UQPE = mc_UQPEhat/iter_MC;
bias_UQPE = mean_UQPE - trueUQPE;
rmse_UQPE = (mc_UQPESqErr/iter_MC)^0.5;
unif_cover80 = mc_boot_uniform_80cover/iter_MC;
unif_cover90 = mc_boot_uniform_90cover/iter_MC;
unif_cover95 = mc_boot_uniform_95cover/iter_MC;
unif_cover99 = mc_boot_uniform_99cover/iter_MC;
point_cover80 = mc_boot_pointwise_80cover/iter_MC;
point_cover90 = mc_boot_pointwise_90cover/iter_MC;
point_cover95 = mc_boot_pointwise_95cover/iter_MC;
point_cover99 = mc_boot_pointwise_99cover/iter_MC;
anal_cover80 = mc_analytic_pointwise_80cover/iter_MC;
anal_cover90 = mc_analytic_pointwise_90cover/iter_MC;
anal_cover95 = mc_analytic_pointwise_95cover/iter_MC;
anal_cover99 = mc_analytic_pointwise_99cover/iter_MC;

ALL_MC_RESULTS = 
list(tau=taulist,
     t_theta=truetheta,m_theta=mean_theta,b_theta=bias_theta,rmse_theta=rmse_theta,
     t_QY=trueQY,m_QY=mean_QY,b_QY=bias_QY,rmse_QY=rmse_QY,
     t_fY=truefY,m_fY=mean_fY,b_fY=bias_fY,rmse_fY=rmse_fY,
     t_UQPE=trueUQPE,m_UQPE=mean_UQPE,b_UQPE=bias_UQPE,rmse_UQPE=rmse_UQPE,
     unif_cover80=unif_cover80,unif_cover90=unif_cover90,unif_cover95=unif_cover95,unif_cover99=unif_cover99,
     point_cover80=point_cover80,point_cover90=point_cover90,point_cover95=point_cover95,point_cover99=point_cover99,
     anal_cover80=anal_cover80,anal_cover90=anal_cover90,anal_cover95=anal_cover95,anal_cover99=anal_cover99)
filename = paste("MC_RESULTS_DGP",DGP,"_N",n,"_P",p,"_L",L,".csv",sep="");
write.csv(ALL_MC_RESULTS,filename);

