
list.of.packages <- c("mvtnorm","glmnet","doParallel")
new.packages <- list.of.packages[!(list.of.packages %in% installed.packages()[,"Package"])]
if(length(new.packages)) install.packages(new.packages)
library(mvtnorm)
library(glmnet)
library(doParallel)
MC_ITER = 500
PARALLEL = 14
DELTA = 0.0000001
TOLERANCE = 0.001

################################################################################
# ESTIMATION PARAMETERS (BEGIN)
################################################################################
taulist = matrix(10:40/50,,1)
b <- function(x){ cbind(x,x^2,x^3) }
h <- function(x){ cbind(x,x^2,x^3,1) }
################################################################################
# ESTIMATION PARAMETERS (ENDED)
################################################################################

################################################################################
# BOOTSTRAP PARAMETERS (BEGIN)
################################################################################
B = 500
################################################################################
# BOOTSTRAP PARAMETERS (ENDED)
################################################################################

################################################################################
# DATA GENERATING PARAMETERS (BEGIN)
################################################################################
p = 100
alpha = matrix(0.5^(2:p/1),p-1,1)
gamma = matrix(0.5^(2:p/1),p-1,1)
Sigma = 0.5^(2*(abs(matrix(1:(p-1),p-1,p-1)-t(matrix(1:(p-1),p-1,p-1)))+1))
DGP = 3
g <- function(x1){ 1*x1 - 0.1*x1^2 + 0.01*x1^3 }
source("_code_approx_trueUQPE.r")
N = 500
################################################################################
# DATA GENERATING PARAMETERS (ENDED)
################################################################################

################################################################################
# BEGIN MONTE CARLO
################################################################################
list_UQPEhat = NULL
list_unif_cover = NULL
list_point_cover = NULL
for( iter in 1:MC_ITER ){

################################################################################
# GENERATE DATA (BEGIN)
################################################################################
X2p = rmvnorm(N, mean=rep(0,nrow(Sigma)), sigma=Sigma)
X1  = matrix(rnorm(N, mean=(X2p%*%gamma), sd=1),N,1)
Y = matrix(rnorm(N, mean=(g(X1)+X2p%*%alpha), sd=1),N,1)
################################################################################
# GENERATE DATA (ENDED)
################################################################################



################################################################################
# ESTIMATE m0 & m1 (BEGIN)
################################################################################
m0hat = matrix(0,N,dim(taulist)[1])
m1hat = matrix(0,N,dim(taulist)[1])
bX = b(cbind(X1,X1^2,X1^3,X2p,X2p^2,X2p^3))
bXsd = apply(bX,2,sd); bXsd[which(bXsd==0)]=1; bXsd = t(matrix(bXsd,dim(bX)[2],N))
bX = bX / bXsd
lambda = 1.1*qnorm(1-0.1/log(N)/max(c(dim(bX)[2],N)))/N^0.5

for( idx in 1:(dim(taulist)[1]) ){
 tau = taulist[idx,1]
 qhattau = quantile(Y,p=tau)

 # LASSO #
 Psihat = colMeans( matrix(Y<=qhattau,N,dim(bX)[2])^2 * bX^2 )^0.5
 lassoresult = glmnet( bX, Y<=qhattau, lambda = lambda, family = "binomial", intercept = TRUE, penalty.factor = Psihat );
 mcheck = cbind(1,bX) %*% coef(lassoresult)
 Psihat = colMeans( matrix(Y<=qhattau-mcheck,N,dim(bX)[2])^2 * bX^2 )^0.5
 lassoresult = glmnet( bX, Y<=qhattau, lambda = lambda, family = "binomial", intercept = TRUE, penalty.factor = Psihat )

 # POST LASSO #
 support = which( as.numeric(coef(lassoresult))[2:(dim(bX)[2]+1)] | 1:(dim(bX)[2])<=3 )
 selected_bX = matrix(bX[,support],N,length(support))
 postlassoresult = glm( (Y<=qhattau) ~ selected_bX, family = "binomial" )
 postlasscoef = matrix(0,dim(bX)[2]+1,1)
 postlasscoef[c(1,support+1)] = coef(postlassoresult)
 postlasscoef[which(is.na(postlasscoef))] = 0

 # m0 & m1 #
 m0hat[,idx] = plogis( cbind(1,bX) %*% postlasscoef )
 m1hat[,idx] = dlogis( cbind(1,bX) %*% postlasscoef ) * cbind(0,(b(cbind(X1+DELTA,(X1+DELTA)^2,(X1+DELTA)^3,X2p,X2p^2,X2p^3))/bXsd-bX)) %*% postlasscoef / DELTA
}
################################################################################
# ESTIMATE m0 & m1 (ENDED)
################################################################################

################################################################################
# ESTIMATE \omega (BEGIN)
################################################################################
# COODINATE DESCEND OBJECTIVE FUNCTION
#subobj <- function(rho1,rho,coord,Mhat,Ghat,lambdaL){ 
# rho[cood,1]=rho1; return( -2*t(Mhat)%*%rho + t(rho)%*%Ghat%*%rho + 2*lambdaL*sum(abs(rho)) ) 
#}
#
#hX = h(cbind(X1,X1^2,X1^3,X2p,X2p^2,X2p^3))
#hXsd = apply(hX,2,sd); hXsd[which(hXsd==0)]=1; hXsd = t(matrix(hXsd,dim(hX)[2],N))
#hX = hX / hXsd
#dimhX = dim(hX)[2]
#Mhat = matrix(-colMeans( (h(cbind(X1+DELTA,(X1+DELTA)^2,(X1+DELTA)^3,X2p,X2p^2,X2p^3)) / hXsd-hX)/DELTA ),,1)
#Ghat = t(hX)%*%hX / N
#
#lambdaL = log(log(N))*(log(dim(hX)[2])/N)^0.5
#rho = matrix(0,dimhX,1); prevrho = rho; DONE = FALSE
#it = 1
#while( !DONE & it < 100 ){
# for( cood in 1:(dim(rho)[1]) ){
#  rho[cood,1] = nlm(subobj,rho[cood,1],rho,cood,Mhat,Ghat,lambdaL)$estimate
# }
# DONE = sum((rho-prevrho)^2)^0.5 < TOLERANCE; prevrho = rho;  it = it + 1;
#}
omegahat = 0 #hX%*%rho
################################################################################
# ESTIMATE \omega (ENDED)
################################################################################

################################################################################
# ESTIMATE \theta (BEGIN)
################################################################################
thetahat = matrix( colMeans( m1hat - matrix(omegahat,N,length(taulist)) * ( ( matrix(Y,N,length(taulist)) <= t(matrix(quantile(Y,p=taulist),length(taulist),N)) ) - m0hat ) ), length(taulist), 1)
################################################################################
# ESTIMATE \theta (ENDED)
################################################################################

################################################################################
# ESTIMATE DENSITY (BEGIN)
################################################################################
K1 <- function(u){ return(0.75*(1-u^2)*(-1 < u)*(u < 1)) }
h1 = 1.06 * sd(Y) / N^0.2 / N^0.05;
fhat = matrix(0,length(taulist),1)
for( idx in 1:length(taulist) ){
 fhat[idx] = mean(K1((Y-quantile(Y,p=taulist[idx]))/h1)/h1)
}
################################################################################
# ESTIMATE DENSITY (ENDED)
################################################################################

################################################################################
# ESTIMATE UQPE (BEGIN)
################################################################################
UQPEhat = - thetahat / fhat
################################################################################
# ESTIMATE UQPE (ENDED)
################################################################################



################################################################################
# BOOTSTRAP BY PARALLEL (BEGIN)
################################################################################
cl<-makeCluster(PARALLEL)
registerDoParallel(cl)
UQPEhatstar = NULL
UQPEhatstar <- foreach( bdx = 1:B, .combine = cbind )%dopar%{
 UQPEhatstar = matrix(0,length(taulist),1)
 library(glmnet);
 eta = matrix(rnorm(N),N,1)

 for( idx in 1:length(taulist) ){
  tau = taulist[idx,1]
  qhattau = quantile(Y,p=tau)
  r_N = floor( 1 + sum( tau + eta*(tau-(Y<=qhattau)) ) )
 
  # q #
  qhattaustar = (Y[order(Y)])[r_N]

  # LASSO #
  Psihat = colMeans( matrix(Y<=qhattaustar,N,dim(bX)[2])^2 * bX^2 )^0.5
  lassoresult = glmnet( bX, Y<=qhattaustar, lambda = lambda, family = "binomial", intercept = TRUE, penalty.factor = Psihat );
  mcheck = cbind(1,bX) %*% coef(lassoresult)
  Psihat = colMeans( matrix(Y<=qhattaustar-mcheck,N,dim(bX)[2])^2 * bX^2 )^0.5
  lassoresult = glmnet( bX, Y<=qhattaustar, lambda = lambda, family = "binomial", intercept = TRUE, penalty.factor = Psihat )

  # POST LASSO #
  support = which( as.numeric(coef(lassoresult))[2:(dim(bX)[2]+1)] | 1:(dim(bX)[2])<=3 )
  selected_bX = matrix(bX[,support],N,length(support))
  postlassoresult = glm( (Y<=qhattaustar) ~ selected_bX, family = "binomial" )
  postlasscoef = matrix(0,dim(bX)[2]+1,1)
  postlasscoef[c(1,support+1)] = coef(postlassoresult);
  postlasscoef[which(is.na(postlasscoef))] = 0

  # m0 & m1 #
  m0hatstar = plogis( cbind(1,bX) %*% postlasscoef )
  m1hatstar = dlogis( cbind(1,bX) %*% postlasscoef ) * cbind(0,(b(cbind(X1+DELTA,(X1+DELTA)^2,(X1+DELTA)^3,X2p,X2p^2,X2p^3))/bXsd-bX)) %*% postlasscoef / DELTA

  thetahatstar = sum( (eta+1) * ( matrix(m1hatstar,,1) - omegahat * ( (Y<=qhattaustar) - matrix(m0hatstar,,1) ) ) ) / sum( eta + 1 )

  # f #
  h1 = 1.06 * sd(Y) / N^0.2;
  fhatstar = sum((eta+1)*K1((Y-qhattaustar)/h1)/h1) / sum(eta+1)

  # UQPE #
  UQPEhatstar[idx,1] = - thetahatstar / fhatstar
 }
 rm(eta);rm(Psihat);rm(lassoresult);rm(mcheck);rm(postlassoresult);rm(m0hatstar);rm(m1hatstar);gc();
 UQPEhatstar
}
stopCluster(cl)
################################################################################
# BOOTSTRAP BY PARALLEL (ENDED)
################################################################################



#rm(X2p);rm(X1);rm(Y);rm(m0hat);rm(m1hat);rm(bX);rm(bXsd);rm(hX);rm(hXsd);rm(Mhat);rm(Ghat);rm(rho);rm(fhat);gc();



################################################################################
# RESULTS (BEGIN)
################################################################################
#sigmahat = matrix(apply(UQPEhatstar,1,sd),length(taulist),1) # Based on Sample SD
sigmahat = matrix((apply(UQPEhatstar,1,quantile,probs=0.75)-apply(UQPEhatstar,1,quantile,probs=0.25))/1.34896,length(taulist),1) # Based on Interquartile Range

point_critvalue = apply(abs(UQPEhatstar - matrix(UQPEhat,length(taulist),B))/matrix(sigmahat,length(taulist),B),1,quantile,probs=0.95)
point_lower = UQPEhat - sigmahat*point_critvalue
point_upper = UQPEhat + sigmahat*point_critvalue
point_cover = (point_lower<=trueUQPE & trueUQPE<=point_upper)

unif_critvalue = quantile( apply(abs(UQPEhatstar - matrix(UQPEhat,length(taulist),B))/matrix(sigmahat,length(taulist),B),2,max), p=0.95 )
unif_lower = UQPEhat - sigmahat*unif_critvalue
unif_upper = UQPEhat + sigmahat*unif_critvalue
unif_cover = min(unif_lower<=trueUQPE & trueUQPE<=unif_upper)
################################################################################
# RESULTS (ENDED)
################################################################################

list_UQPEhat = rbind(list_UQPEhat,t(UQPEhat))
list_unif_cover = rbind(list_unif_cover,unif_cover)
list_point_cover = rbind(list_point_cover,t(point_cover))

print(c(iter,colMeans(list_point_cover)[c(1,11,21,31)],0,mean(list_unif_cover))); Sys.sleep(0.00001); flush.console();
}
################################################################################
# END MONTE CARLO
################################################################################



MCmean = matrix(colMeans(list_UQPEhat),,1)
MCbias = MCmean - trueUQPE
MCRMSE = (MCbias^2 + matrix(apply(list_UQPEhat,2,var),,1))^0.5
MCpcov = matrix(colMeans(list_point_cover),,1)
MCucov = mean(list_unif_cover)

RESULTS = cbind(taulist,trueUQPE,MCmean,MCbias,MCRMSE,MCpcov,MCucov)
colnames(RESULTS) <- c("Tau","True","Mean","Bias","RMSE","Point95","Unif95")
RESULTS 
write.csv(RESULTS[c(1,11,21,31),],paste("MC_DGP",DGP,"_N",N,"_p",p,".csv",sep=""))