################################################################################
# INSTALL PACKAGES AND CALL LIBRARIES
################################################################################
list.of.packages <- c("mvtnorm","glmnet","doParallel")
new.packages <- list.of.packages[!(list.of.packages %in% installed.packages()[,"Package"])]
if(length(new.packages)) install.packages(new.packages)
library(glmnet);
library(hqreg);
library(doParallel)

################################################################################
# BASIC SETTINGS
################################################################################
DELTA = 0.0000001
TOLERANCE = 0.001
taulist = matrix(10:40/50,,1)
B = 500
PARALLEL = 4

################################################################################
# DATA
################################################################################
load("data.RData");
head(y); # days_per_week hours_per_day hours_per_week hourly_wage weekly_earn
head(d); # jc_days class_days
y = matrix(y[,"hourly_wage"],,1)
d1 = matrix(d[,"jc_days"],,1)
d2 = matrix(d[,"class_days"],,1)
data <- cbind(y,d1,d2,x);
data <- na.omit(data);
N = dim(data)[1]; 
p = dim(data)[2]-3;
Y = matrix(data[,1],,1);
#Y = log(Y);
X1 = matrix(data[,2],,1);   # '2' for "jc_days" and '3' for "class_days"
X2p = as.matrix(data[,4:(dim(data)[2])])

for( idx in 1:(dim(X2p)[2]) ){
 if( sd(X2p[,idx])>0 ){ 
  X2p[,idx] = (X2p[,idx]-mean(X2p[,idx]))/sd(X2p[,idx])
 }
}
p = dim(X2p)[2]

################################################################################
# ESTIMATION PARAMETERS (BEGIN)
################################################################################
taulist = matrix(10:40/50,,1)
b <- function(x){ cbind(x,x^2,x^3) }
h <- function(x){ cbind(x,x^2,x^3,1) }
################################################################################
# ESTIMATION PARAMETERS (ENDED)
################################################################################

################################################################################
# ESTIMATE m0 & m1 (BEGIN)
################################################################################
m0hat = matrix(0,N,dim(taulist)[1])
m1hat = matrix(0,N,dim(taulist)[1])
bX = b(cbind(X1,X1^2,X1^3,X2p,X2p^2,X2p^3))
bXsd = apply(bX,2,sd); bXsd[which(bXsd==0)]=1; bXsd = t(matrix(bXsd,dim(bX)[2],N))
bX = bX / bXsd
lambda = 1.1*qnorm(1-0.1/log(N)/max(c(dim(bX)[2],N)))/N^0.5

for( idx in 1:(dim(taulist)[1]) ){
 tau = taulist[idx,1]
 qhattau = quantile(Y,p=tau)

 # LASSO #
 Psihat = colMeans( matrix(Y<=qhattau,N,dim(bX)[2])^2 * bX^2 )^0.5 + 0.00001
 lassoresult = glmnet( bX, Y<=qhattau, lambda = lambda, family = "binomial", intercept = TRUE, penalty.factor = Psihat );
 mcheck = cbind(1,bX) %*% coef(lassoresult)
 Psihat = colMeans( matrix(Y<=qhattau-mcheck,N,dim(bX)[2])^2 * bX^2 )^0.5 + 0.00001
 lassoresult = glmnet( bX, Y<=qhattau, lambda = lambda, family = "binomial", intercept = TRUE, penalty.factor = Psihat )

 # POST LASSO #
 support = which( as.numeric(coef(lassoresult))[2:(dim(bX)[2]+1)] | 1:(dim(bX)[2])<=3 )
 selected_bX = matrix(bX[,support],N,length(support))
 postlassoresult = glm( (Y<=qhattau) ~ selected_bX, family = "binomial" )
 postlasscoef = matrix(0,dim(bX)[2]+1,1)
 postlasscoef[c(1,support+1)] = coef(postlassoresult)
 postlasscoef[which(is.na(postlasscoef))] = 0

 # m0 & m1 #
 m0hat[,idx] = plogis( cbind(1,bX) %*% postlasscoef )
 m1hat[,idx] = dlogis( cbind(1,bX) %*% postlasscoef ) * cbind(0,(b(cbind(X1+DELTA,(X1+DELTA)^2,(X1+DELTA)^3,X2p,X2p^2,X2p^3))/bXsd-bX)) %*% postlasscoef / DELTA
}
################################################################################
# ESTIMATE m0 & m1 (ENDED)
################################################################################

################################################################################
# ESTIMATE \omega (BEGIN)
################################################################################
# COODINATE DESCEND OBJECTIVE FUNCTION
subobj <- function(rho1,rho,coord,Mhat,Ghat,lambdaL){ 
 rho[cood,1]=rho1; return( -2*t(Mhat)%*%rho + t(rho)%*%Ghat%*%rho + 2*lambdaL*sum(abs(rho)) ) 
}

hX = h(cbind(X1,X1^2,X1^3,X2p,X2p^2,X2p^3))
hXsd = apply(hX,2,sd); hXsd[which(hXsd==0)]=1; hXsd = t(matrix(hXsd,dim(hX)[2],N))
hX = hX / hXsd
dimhX = dim(hX)[2]
Mhat = matrix(-colMeans( (h(cbind(X1+DELTA,(X1+DELTA)^2,(X1+DELTA)^3,X2p,X2p^2,X2p^3)) / hXsd-hX)/DELTA ),,1)
Ghat = t(hX)%*%hX / N

lambdaL = log(log(N))*(log(dim(hX)[2])/N)^0.5
rho = matrix(0,dimhX,1); prevrho = rho; DONE = FALSE
it = 1
while( !DONE & it < 100 ){
 for( cood in 1:(dim(rho)[1]) ){
  rho[cood,1] = nlm(subobj,rho[cood,1],rho,cood,Mhat,Ghat,lambdaL)$estimate
 }
 DONE = sum((rho-prevrho)^2)^0.5 < TOLERANCE; prevrho = rho;  it = it + 1;
}
omegahat = hX%*%rho
################################################################################
# ESTIMATE \omega (ENDED)
################################################################################

################################################################################
# ESTIMATE \theta (BEGIN)
################################################################################
thetahat = matrix( colMeans( m1hat - matrix(omegahat,N,length(taulist)) * ( ( matrix(Y,N,length(taulist)) <= t(matrix(quantile(Y,p=taulist),length(taulist),N)) ) - m0hat ) ), length(taulist), 1)
################################################################################
# ESTIMATE \theta (ENDED)
################################################################################

################################################################################
# ESTIMATE DENSITY (BEGIN)
################################################################################
K1 <- function(u){ return(0.75*(1-u^2)*(-1 < u)*(u < 1)) }
h1 = 1.06 * sd(Y) / N^0.2 / N^0.01;
fhat = matrix(0,length(taulist),1)
for( idx in 1:length(taulist) ){
 fhat[idx] = mean(K1((Y-quantile(Y,p=taulist[idx]))/h1)/h1)
}
################################################################################
# ESTIMATE DENSITY (ENDED)
################################################################################

################################################################################
# ESTIMATE UQPE (BEGIN)
################################################################################
UQPEhat = - thetahat / fhat
################################################################################
# ESTIMATE UQPE (ENDED)
################################################################################

################################################################################
# BOOTSTRAP BY PARALLEL FOR THETA (BEGIN)
################################################################################
cl<-makeCluster(PARALLEL)
registerDoParallel(cl)
thetahatstar = NULL
thetahatstar <- foreach( bdx = 1:B, .combine = cbind )%dopar%{
 thetahatstar = matrix(0,length(taulist),1)
 library(glmnet);
 eta = matrix(rnorm(N),N,1)

 for( idx in 1:length(taulist) ){
  tau = taulist[idx,1]
  qhattau = quantile(Y,p=tau)
  r_N = floor( 1 + sum( tau + eta*(tau-(Y<=qhattau)) ) )
 
  # q #
  qhattaustar = (Y[order(Y)])[r_N]

  # LASSO #
  Psihat = colMeans( matrix(Y<=qhattaustar,N,dim(bX)[2])^2 * bX^2 )^0.5 + 0.00001
  lassoresult = glmnet( bX, Y<=qhattaustar, lambda = lambda, family = "binomial", intercept = TRUE, penalty.factor = Psihat );
  mcheck = cbind(1,bX) %*% coef(lassoresult)
  Psihat = colMeans( matrix(Y<=qhattaustar-mcheck,N,dim(bX)[2])^2 * bX^2 )^0.5 + 0.00001
  lassoresult = glmnet( bX, Y<=qhattaustar, lambda = lambda, family = "binomial", intercept = TRUE, penalty.factor = Psihat )

  # POST LASSO #
  support = which( as.numeric(coef(lassoresult))[2:(dim(bX)[2]+1)] | 1:(dim(bX)[2])<=3 )
  selected_bX = matrix(bX[,support],N,length(support))
  postlassoresult = glm( (Y<=qhattaustar) ~ selected_bX, family = "binomial" )
  postlasscoef = matrix(0,dim(bX)[2]+1,1)
  postlasscoef[c(1,support+1)] = coef(postlassoresult);
  postlasscoef[which(is.na(postlasscoef))] = 0

  # m0 & m1 #
  m0hatstar = plogis( cbind(1,bX) %*% postlasscoef )
  m1hatstar = dlogis( cbind(1,bX) %*% postlasscoef ) * cbind(0,(b(cbind(X1+DELTA,(X1+DELTA)^2,(X1+DELTA)^3,X2p,X2p^2,X2p^3))/bXsd-bX)) %*% postlasscoef / DELTA

  thetahatstar[idx,1] = sum( (eta+1) * ( matrix(m1hatstar,,1) - omegahat * ( (Y<=qhattaustar) - matrix(m0hatstar,,1) ) ) ) / sum( eta + 1 )
 }
 rm(eta);rm(Psihat);rm(lassoresult);rm(mcheck);rm(postlassoresult);rm(m0hatstar);rm(m1hatstar);gc();
 thetahatstar
}
stopCluster(cl)
################################################################################
# BOOTSTRAP BY PARALLEL FOR THETA (ENDED)
################################################################################

################################################################################
# BOOTSTRAP BY PARALLEL FOR UQPE (BEGIN)
################################################################################
cl<-makeCluster(PARALLEL)
registerDoParallel(cl)
UQPEhatstar = NULL
UQPEhatstar <- foreach( bdx = 1:B, .combine = cbind )%dopar%{
 UQPEhatstar = matrix(0,length(taulist),1)
 library(glmnet);
 eta = matrix(rnorm(N),N,1)

 for( idx in 1:length(taulist) ){
  tau = taulist[idx,1]
  qhattau = quantile(Y,p=tau)
  r_N = floor( 1 + sum( tau + eta*(tau-(Y<=qhattau)) ) )
 
  # q #
  qhattaustar = (Y[order(Y)])[r_N]

  # LASSO #
  Psihat = colMeans( matrix(Y<=qhattaustar,N,dim(bX)[2])^2 * bX^2 )^0.5 + 0.00001
  lassoresult = glmnet( bX, Y<=qhattaustar, lambda = lambda, family = "binomial", intercept = TRUE, penalty.factor = Psihat );
  mcheck = cbind(1,bX) %*% coef(lassoresult)
  Psihat = colMeans( matrix(Y<=qhattaustar-mcheck,N,dim(bX)[2])^2 * bX^2 )^0.5 + 0.00001
  lassoresult = glmnet( bX, Y<=qhattaustar, lambda = lambda, family = "binomial", intercept = TRUE, penalty.factor = Psihat )

  # POST LASSO #
  support = which( as.numeric(coef(lassoresult))[2:(dim(bX)[2]+1)] | 1:(dim(bX)[2])<=3 )
  selected_bX = matrix(bX[,support],N,length(support))
  postlassoresult = glm( (Y<=qhattaustar) ~ selected_bX, family = "binomial" )
  postlasscoef = matrix(0,dim(bX)[2]+1,1)
  postlasscoef[c(1,support+1)] = coef(postlassoresult);
  postlasscoef[which(is.na(postlasscoef))] = 0

  # m0 & m1 #
  m0hatstar = plogis( cbind(1,bX) %*% postlasscoef )
  m1hatstar = dlogis( cbind(1,bX) %*% postlasscoef ) * cbind(0,(b(cbind(X1+DELTA,(X1+DELTA)^2,(X1+DELTA)^3,X2p,X2p^2,X2p^3))/bXsd-bX)) %*% postlasscoef / DELTA

  thetahatstar = sum( (eta+1) * ( matrix(m1hatstar,,1) - omegahat * ( (Y<=qhattaustar) - matrix(m0hatstar,,1) ) ) ) / sum( eta + 1 )

  # f #
  h1 = 1.06 * sd(Y) / N^0.2;
  fhatstar = sum((eta+1)*K1((Y-qhattaustar)/h1)/h1) / sum(eta+1)

  # UQPE #
  UQPEhatstar[idx,1] = - thetahatstar / fhatstar
 }
 rm(eta);rm(Psihat);rm(lassoresult);rm(mcheck);rm(postlassoresult);rm(m0hatstar);rm(m1hatstar);gc();
 UQPEhatstar
}
stopCluster(cl)
################################################################################
# BOOTSTRAP BY PARALLEL FOR UQPE (ENDED)
################################################################################

################################################################################
# RESULTS (BEGIN)
################################################################################
sigmahattheta = matrix((apply(thetahatstar,1,quantile,probs=0.75)-apply(thetahatstar,1,quantile,probs=0.25))/1.34896,length(taulist),1) # Based on Interquartile Range
sigmahat = matrix((apply(UQPEhatstar,1,quantile,probs=0.75)-apply(UQPEhatstar,1,quantile,probs=0.25))/1.34896,length(taulist),1) # Based on Interquartile Range

point_critvalue = apply(abs(UQPEhatstar - matrix(UQPEhat,length(taulist),B))/matrix(sigmahat,length(taulist),B),1,quantile,probs=0.95)
point_lower = UQPEhat - sigmahat*point_critvalue
point_upper = UQPEhat + sigmahat*point_critvalue

unif_critvalue = quantile( apply(abs(UQPEhatstar - matrix(UQPEhat,length(taulist),B))/matrix(sigmahat,length(taulist),B),2,max), p=0.95 )
unif_lower = UQPEhat - sigmahat*unif_critvalue
unif_upper = UQPEhat + sigmahat*unif_critvalue

sigmahattheta = matrix((apply(thetahatstar,1,quantile,probs=0.75)-apply(thetahatstar,1,quantile,probs=0.25))/1.34896,length(taulist),1) # Based on Interquartile Range

theta_unif_critvalue = quantile( apply(abs(thetahatstar - matrix(thetahat,length(taulist),B))/matrix(sigmahattheta,length(taulist),B),2,max), p=0.95 )
theta_unif_lower = thetahat - sigmahattheta*theta_unif_critvalue
theta_unif_upper = thetahat + sigmahattheta*theta_unif_critvalue
################################################################################
# RESULTS (ENDED)
################################################################################

tauselect=c(1,6,11,16,21,26,31)
ylim=1000*c(min(unif_lower[tauselect,]),max(unif_upper[tauselect,]))
plot(taulist[tauselect,], 1000*UQPEhat[tauselect,], type="l", ylim=ylim, xlab=expression(tau), ylab=expression(paste("1000 X UQPE(",tau,")")))
par(new=TRUE)
plot(taulist[tauselect,], 1000*point_lower[tauselect,], type="l", lty=2, ylim=ylim, xlab="", ylab="")
par(new=TRUE)
plot(taulist[tauselect,], 1000*point_upper[tauselect,], type="l", lty=2, ylim=ylim, xlab="", ylab="")
par(new=TRUE)
plot(taulist[tauselect,], 1000*unif_lower[tauselect,], type="l", lty=3, ylim=ylim, xlab="", ylab="")
par(new=TRUE)
plot(taulist[tauselect,], 1000*unif_upper[tauselect,], type="l", lty=3, ylim=ylim, xlab="", ylab="")
abline(h=0, col="gray")
legend(min(taulist),max(ylim),c("Estimates","Pointwise CI","Uniform CB"),lty=1:3)
cbind(taulist, 1000*UQPEhat, 1000*point_lower, 1000*point_upper)[c(1,11,21,31),]
cbind(taulist, 1000*UQPEhat, 1000*unif_lower, 1000*unif_upper)[c(1,11,21,31),]

reject_uniform_0 = sum(theta_unif_lower > 0 | 0 < theta_unif_upper)>0
reject_uniform_0